"""Pydaikin appliance, represent a Daikin AirBase device."""

import logging
from urllib.parse import quote, unquote

from .daikin_brp069 import DaikinBRP069

_LOGGER = logging.getLogger(__name__)


class DaikinAirBase(DaikinBRP069):
    """Daikin class for AirBase (BRP15B61) units."""

    TRANSLATIONS = dict(
        DaikinBRP069.TRANSLATIONS,
        **{
            'mode': {
                '0': 'fan',
                '1': 'hot',
                '2': 'cool',
                '3': 'auto',
                '7': 'dry',
            },
            'f_rate': {
                '1': 'low',
                '3': 'mid',
                '5': 'high',
                '1a': 'low/auto',
                '3a': 'mid/auto',
                '5a': 'high/auto',
            },
        },
    )

    HTTP_RESOURCES = [
        'common/basic_info',
        'aircon/get_control_info',
        'aircon/get_model_info',
        'aircon/get_sensor_info',
        'aircon/get_zone_setting',
    ]

    INFO_RESOURCES = DaikinBRP069.INFO_RESOURCES + ['aircon/get_zone_setting']

    @staticmethod
    def parse_response(response_body):
        """Parse response from Daikin, add support for f_rate-auto."""
        _LOGGER.debug("Parsing %s", response_body)
        response = super(DaikinAirBase, DaikinAirBase).parse_response(response_body)
        if response.get('f_auto') == '1':
            response['f_rate'] = f'{response["f_rate"]}a'
        return response

    def __init__(self, device_id, session=None):
        """Init the pydaikin appliance, representing one Daikin AirBase
        (BRP15B61) device."""
        super().__init__(device_id, session)
        self.values.update({'htemp': '-', 'otemp': '-', 'shum': '--'})

    async def init(self):
        """Init status and set defaults."""
        await super().init()
        if not self.values:
            raise Exception("Empty values.")

    async def _run_get_resource(self, resource):
        """Make the http request."""
        resource = 'skyfi/%s' % resource
        return await super()._run_get_resource(resource)

    @property
    def support_away_mode(self):
        """Return True if the device support away_mode."""
        return False

    @property
    def support_swing_mode(self):
        """Return True if the device support setting swing_mode."""
        return False

    @property
    def support_outside_temperature(self):
        """AirBase unit returns otemp if master controller starts before it."""
        return True

    @property
    def fan_rate(self):
        """Return list of supported fan rates."""
        fan_rates = list(map(str.title, self.TRANSLATIONS.get('f_rate', {}).values()))
        if self.values.get('frate_steps') == '2':
            if self.values.get('en_frate_auto') == '0':
                return fan_rates[:3:2]
            return fan_rates[:3:2] + fan_rates[3::2]
        if self.values.get('en_frate_auto') == '0':
            return fan_rates[:3]
        return fan_rates

    async def _update_settings(self, settings):
        """Update settings to set on Daikin device."""

        # Call the base BRP069 method to update the settings; it will
        # return the current values it retrieves from the controller
        # so we can further process them
        current_val = await super()._update_settings(settings)

        # f_auto requires some special handling, as it is managed as an
        # attribute of f_rate and we don't directly set it - so when f_rate
        # is being changed, ensure we update f_auto accordingly if it is
        # defined in the current device's returned settings
        if 'f_auto' in current_val:
            # The system supports f_auto; if we are setting the fan speed
            # then ensure we update the f_auto setting as well
            if 'f_rate' in settings:
                self.values['f_auto'] = '1' if 'a' in self.values['f_rate'] else '0'
            else:
                key = 'auto' + self.values['mode']
                if key in current_val:
                    self.values['f_auto'] = current_val[key]

                    # The f_rate value would have been retrieved from the unit's current
                    # operating mode fan rate setting, and needs the 'a' suffix reinstated
                    # if we are running in an automatic fan speed mode
                    if self.values['f_auto'] == '1':
                        self.values['f_rate'] = f'{self.values["f_rate"]}a'

        return current_val

    async def set(self, settings):
        """Set settings on Daikin device."""
        await self._update_settings(settings)

        self.values.setdefault('f_airside', 0)
        query_c = (
            'aircon/set_control_info'
            '?pow={pow}&mode={mode}&stemp={stemp}&shum={shum}'
            '&f_rate={f_rate[0]}&f_auto={f_auto}&f_dir={f_dir}'
            '&lpw=&f_airside={f_airside}'
        ).format(**self.values)

        _LOGGER.debug("Sending query_c: %s", query_c)
        await self._get_resource(query_c)

    def represent(self, key):
        """Return translated value from key."""
        k, val = super().represent(key)

        if key in ['zone_name', 'zone_onoff']:
            val = unquote(self.values[key]).split(';')

        return (k, val)

    @property
    def zones(self):
        """Return list of zones."""
        if not self.values.get('zone_name'):
            return None
        zone_onoff = self.represent('zone_onoff')[1]
        return [
            (name.strip(' +,'), zone_onoff[i])
            for i, name in enumerate(self.represent('zone_name')[1])
        ]

    async def set_zone(self, zone_id, status):
        """Set zone status."""
        current_state = await self._get_resource('aircon/get_zone_setting')
        self.values.update(current_state)
        zone_onoff = self.represent('zone_onoff')[1]
        zone_onoff[zone_id] = status
        self.values['zone_onoff'] = quote(';'.join(zone_onoff)).lower()

        query = 'aircon/set_zone_setting?zone_name={}&zone_onoff={}'.format(
            current_state['zone_name'],
            self.values['zone_onoff'],
        )

        _LOGGER.debug("Set zone:: %s", query)
        await self._get_resource(query)



"""Discovery module to autodiscover Daikin devices on local network."""

import logging
import socket

import netifaces

_LOGGER = logging.getLogger(__name__)

UDP_SRC_PORT = 30000
UDP_DST_PORT = 30050
RCV_BUFSIZ = 1024

GRACE_SECONDS = 1

DISCOVERY_MSG = "DAIKIN_UDP/common/basic_info"


class DiscoveredObject:
    """Class to represent discovered object."""

    def __init__(self, ip, port, basic_info_string):
        self.values = {}

        self.values['ip'] = ip
        self.values['port'] = port
        self.values.update(self.parse_basic_info(basic_info_string))

    @staticmethod
    def parse_basic_info(basic_info):
        """Parse basic info."""
        from .daikin_base import Appliance  # pylint: disable=import-outside-toplevel

        data = Appliance.parse_response(basic_info)

        if 'mac' not in data:
            raise ValueError("no mac found for device")

        return data

    def __getitem__(self, name):
        if name in self.values:
            return self.values[name]
        raise AttributeError("No such attribute: " + name)

    def keys(self):
        """Returns keys."""
        return self.values.keys()

    def __str__(self):
        return str(self.values)


class Discovery:  # pylint: disable=too-few-public-methods
    """Discovery class."""

    def __init__(self):
        sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        sock.setsockopt(socket.SOL_SOCKET, socket.SO_BROADCAST, 1)
        sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        sock.bind(("", UDP_SRC_PORT))
        sock.settimeout(GRACE_SECONDS)

        self.sock = sock
        self.dev = {}

    def poll(self, stop_if_found=None, ip=None):  # pylint: disable=invalid-name
        """Poll for available devices."""
        if ip:
            broadcast_ips = [ip]
        else:
            # get all IPv4 definitions in the system
            net_groups = [
                netifaces.ifaddresses(i)[netifaces.AF_INET]
                for i in netifaces.interfaces()
                if netifaces.AF_INET in netifaces.ifaddresses(i)
            ]

            # flatten the previous list
            net_ips = [item for sublist in net_groups for item in sublist]

            # from those, get the broadcast IPs, if available
            broadcast_ips = [i['broadcast'] for i in net_ips if 'broadcast' in i.keys()]

        # send a daikin broadcast to each one of the ips
        for ip_address in broadcast_ips:
            self.sock.sendto(bytes(DISCOVERY_MSG, 'UTF-8'), (ip_address, UDP_DST_PORT))

        try:
            while True:  # for anyone who ansers
                data, addr = self.sock.recvfrom(RCV_BUFSIZ)
                _LOGGER.debug("Discovered %s, %s", addr, data.decode('UTF-8'))

                try:
                    data = DiscoveredObject(addr[0], addr[1], data.decode('UTF-8'))

                    new_mac = data['mac']
                    self.dev[new_mac] = data

                    if (
                        stop_if_found is not None
                        and data['name'].lower() == stop_if_found.lower()
                    ):
                        return self.dev.values()

                except ValueError:  # invalid message received
                    continue

        except socket.timeout:  # nobody else is answering
            pass

        return self.dev.values()


def get_devices():
    """Returns discovered devices."""
    discovery = Discovery()

    return discovery.poll()


def get_name(name):
    """Returns the name of discovered devices."""
    disovery = Discovery()

    devices = disovery.poll(name)

    ret = None

    for device in devices:

        if device['name'].lower() == name.lower():
            ret = device

    return ret



"""Pydaikin power mixin."""

from collections import namedtuple
from datetime import datetime, timedelta
import logging

ENERGY_CONSUMPTION_MAX_HISTORY = timedelta(hours=6)

ATTR_TOTAL = 'total'
ATTR_COOL = 'cool'
ATTR_HEAT = 'heat'

TIME_TODAY = 'today'
TIME_YESTERDAY = 'yesterday'
TIME_THIS_YEAR = 'this_year'
TIME_LAST_YEAR = 'last_year'

EnergyConsumptionParser = namedtuple(
    'EnergyConsumptionParser', ['dimension', 'reducer', 'divider']
)

EnergyConsumptionState = namedtuple(
    'EnergyConsumptionState', ['datetime', 'first_state', 'today', 'yesterday']
)

_LOGGER = logging.getLogger(__name__)


class DaikinPowerMixin:
    """Mixin to provide power monitoring capability"""

    _energy_consumption_history = None
    values = None

    ENERGY_CONSUMPTION_PARSERS = {
        f'{ATTR_TOTAL}_{TIME_TODAY}': EnergyConsumptionParser(
            dimension='datas', reducer=lambda values: values[-1], divider=1000
        ),
        f'{ATTR_COOL}_{TIME_TODAY}': EnergyConsumptionParser(
            dimension='curr_day_cool', reducer=sum, divider=10
        ),
        f'{ATTR_HEAT}_{TIME_TODAY}': EnergyConsumptionParser(
            dimension='curr_day_heat', reducer=sum, divider=10
        ),
        f'{ATTR_TOTAL}_{TIME_YESTERDAY}': EnergyConsumptionParser(
            dimension='datas', reducer=lambda values: values[-2], divider=1000
        ),
        f'{ATTR_COOL}_{TIME_YESTERDAY}': EnergyConsumptionParser(
            dimension='prev_1day_cool', reducer=sum, divider=10
        ),
        f'{ATTR_HEAT}_{TIME_YESTERDAY}': EnergyConsumptionParser(
            dimension='prev_1day_heat', reducer=sum, divider=10
        ),
        f'{ATTR_TOTAL}_{TIME_THIS_YEAR}': EnergyConsumptionParser(
            dimension='this_year', reducer=sum, divider=1
        ),
        f'{ATTR_TOTAL}_{TIME_LAST_YEAR}': EnergyConsumptionParser(
            dimension='previous_year', reducer=sum, divider=1
        ),
    }

    @property
    def support_energy_consumption(self):
        """Return True if the device supports energy consumption monitoring."""
        return (self.energy_consumption(mode=ATTR_TOTAL, time=TIME_THIS_YEAR) or 0) + (
            self.energy_consumption(mode=ATTR_TOTAL, time=TIME_LAST_YEAR) or 0
        ) > 0

    def _register_energy_consumption_history(self):
        if not self.support_energy_consumption:
            return

        for mode in (ATTR_TOTAL, ATTR_COOL, ATTR_HEAT):
            new_state = EnergyConsumptionState(
                datetime=datetime.utcnow(),
                first_state=not (self._energy_consumption_history[mode]),
                today=self.energy_consumption(mode=mode, time=TIME_TODAY),
                yesterday=self.energy_consumption(mode=mode, time=TIME_YESTERDAY),
            )
            if new_state.today is None:
                continue

            if not new_state.first_state:
                old_state = self._energy_consumption_history[mode][0]

                if new_state.today == old_state.today:
                    if new_state.yesterday == old_state.yesterday:
                        # State has not changed, nothing to register,
                        # we just update the cmp_freq average
                        continue

            self._energy_consumption_history[mode].insert(0, new_state)

            # We can remove very old states (except the latest one)
            idx = (
                min(
                    (
                        i
                        for i, state in enumerate(
                            self._energy_consumption_history[mode]
                        )
                        if state.datetime
                        < datetime.utcnow() - ENERGY_CONSUMPTION_MAX_HISTORY
                    ),
                    default=len(self._energy_consumption_history[mode]),
                )
                + 1
            )

            self._energy_consumption_history[mode] = self._energy_consumption_history[
                mode
            ][:idx]

    def energy_consumption(self, mode=ATTR_TOTAL, time=TIME_TODAY):
        """Return today/yesterday energy consumption in kWh of a given mode."""
        parser = self.ENERGY_CONSUMPTION_PARSERS.get(f'{mode}_{time}')
        if parser is None:
            raise ValueError(f'Unsupported mode {mode} on {time}.')

        try:
            values = [int(x) for x in self.values.get(parser.dimension).split('/')]
            value = parser.reducer(values)
            value /= parser.divider
            return value
        except (TypeError, IndexError, AttributeError, ValueError):
            return None

    @staticmethod
    def _compute_diff_energy(mode: str, curr, prev):
        """Return the energy consumption delta between two states"""
        if curr.today > prev.today:
            # Normal behavior, today state is growing
            return curr.today - prev.today

        if curr.yesterday is None:
            _LOGGER.error(
                'Decreasing today state and missing yesterday state caused an '
                'impossible energy consumption measure of %s',
                mode,
            )
            return None

        if curr.yesterday >= prev.today:
            # If today state is not growing (or even declines), we probably have
            # shifted 1 day. Thus we should have yesterday state greater or equal
            # to previous today state (in most cases it will be equal)
            return curr.yesterday - prev.today + curr.today

        _LOGGER.error('Impossible energy consumption measure of %s', mode)
        return None

    def current_power_consumption(  # pylint: disable=too-many-branches
        self,
        mode=ATTR_TOTAL,
        exp_diff_time_value=None,
        exp_diff_time_margin_factor=None,
        min_power=0.1,
    ):
        """
        Return the current power consumption of a given mode by estimating the slope
        of the energy consumption. When 100Wh have been consumed, it is assumed that
        the next 100Wh will be consumed in the same duration with a given margin in
        case the power consumption has been lowered (to smooth the consumption).
        """
        if exp_diff_time_value is None and exp_diff_time_margin_factor is None:
            exp_diff_time_margin_factor = timedelta(minutes=5)

        if exp_diff_time_value is not None and not isinstance(
            exp_diff_time_value, timedelta
        ):
            raise TypeError(exp_diff_time_value)
        if exp_diff_time_margin_factor is not None and not isinstance(
            exp_diff_time_margin_factor, (timedelta, float)
        ):
            raise TypeError(exp_diff_time_margin_factor)

        if not self._energy_consumption_history:
            # The sensor has not been properly initialized
            return 0

        history = list(reversed(self._energy_consumption_history[mode]))

        energy_to_log = 0
        exp_diff_time = None
        est_power = 0

        for prev, curr in zip(history, history[1:]):
            diff_time = curr.datetime - prev.datetime
            diff_energy = self._compute_diff_energy(mode, curr, prev)

            # We remove the energy we've logged since last state update
            # This is to fix an incorrect estimation of the previous exp_diff_time
            if exp_diff_time and est_power > 0:
                # We know that the power will be cut off once the exp_diff_time is
                # surpassed. Note this can result in negative value of energy_to_log
                # when the exp_diff_timehas been over-estimated.
                energy_to_log -= max(est_power, min_power) * (
                    min(exp_diff_time, diff_time).total_seconds() / 3600
                )

            # We expect the consumption to be stable so the next diff_time should be
            # barely the same as the previous one. If we over-estimate this duration,
            # it will result in an irregular power consumption, often going back to 0.
            # If we under-estimate this duration, it will ultimately result in a too
            # smoothed power consumption.  Feel free to fine-tune this variable to fit
            # your needs...
            if exp_diff_time_value is None:
                if prev.first_state:
                    # We skip the first state as we cannot trust its datetime for
                    # exp_diff_time estimation
                    continue
                exp_diff_time = diff_time
            else:
                exp_diff_time = exp_diff_time_value

            # Once we have estimated the next diff_time we can compute the estimated
            # current power
            if diff_energy is not None:
                energy_to_log += diff_energy
            est_power = energy_to_log / (exp_diff_time.total_seconds() / 3600)
            est_power = max(est_power, 0)

            # We add some margins to the exp_diff_time AFTER the est_power computation
            # We prefer having an accurate est_power than an accurate est_energy
            if isinstance(exp_diff_time_margin_factor, timedelta):
                exp_diff_time += exp_diff_time_margin_factor
            if isinstance(exp_diff_time_margin_factor, float):
                exp_diff_time *= 1 + exp_diff_time_margin_factor

            if min_power is not None and est_power > 0:
                est_power = max(est_power, min_power)

        if exp_diff_time and datetime.utcnow() > history[-1].datetime + exp_diff_time:
            # The power estimation was computed for a given duration
            # So if we exceed this duration we should return a zero power
            est_power = 0

        if min_power is not None and est_power > 0:
            est_power = max(est_power, min_power)

        return est_power
